from pg_alchemy_kit.PGUtils import PGUtils, get_engine, get_engine_url
from pg_alchemy_kit.PGUtilsORM import PGUtilsORM
from pg_alchemy_kit.PGUtilsBase import PGUtilsBase

from sqlalchemy.orm.session import Session
from sqlalchemy import inspect
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm import DeclarativeMeta
import sqlalchemy
import logging
from contextlib import contextmanager
from typing import List, Iterator

from sqlalchemy import orm


class PG:
    def initialize(
        cls,
        url: str = None,
        logger: logging.Logger = None,
        single_transaction: bool = False,
        pgUtils: PGUtilsBase = PGUtils,
        **kwargs,
    ):
        pg_utils_kwargs = kwargs.pop("pg_utils_kwargs", {})
        session_maker_kwargs = kwargs.pop("session_maker_kwargs", {})

        url = url or get_engine_url()
        cls.engine: Engine = get_engine(url, **kwargs)
        cls.SessionLocal = sessionmaker(
            autocommit=False, autoflush=False, bind=cls.engine, **session_maker_kwargs
        )
        cls.inspector = inspect(cls.engine)
        cls.logger = logger

        if cls.logger is None:
            logger = logging.getLogger(__name__)
            logger.setLevel(logging.INFO)
            logger.addHandler(logging.StreamHandler())
            cls.logger = logger

        cls.utils: PGUtilsBase = pgUtils(
            cls.logger, single_transaction, **pg_utils_kwargs
        )

        cls.logger.info("Initialized PG")

    def create_tables(
        cls, Bases: List[DeclarativeMeta], schemas: List[str] = ["public"]
    ):
        """
        Creates tables for all the models in the list of Bases
        """
        if type(Bases) != list:
            Bases = [Bases]

        if type(schemas) != list:
            schemas = [schemas]

        with cls.engine.begin() as conn:
            for Base, schema in zip(Bases, schemas):
                try:
                    if schema not in cls.inspector.get_schema_names():
                        conn.execute(sqlalchemy.schema.CreateSchema(schema))
                    Base.metadata.create_all(cls.engine)
                except Exception as e:
                    cls.logger.info(f"Error in create_tables: {e}")

    @contextmanager
    def get_session_ctx(cls) -> Iterator[Session]:
        with cls.SessionLocal() as session:
            try:
                cls.utils.initialize(session)
                yield session
            finally:
                session.close()

    @contextmanager
    def transaction(cls) -> Iterator[Session]:
        with cls.SessionLocal() as session:
            try:
                cls.utils.initialize(session)
                yield session
                session.commit()
            except Exception as e:
                session.rollback()
                raise e
            finally:
                session.close()

    def get_session(cls) -> Iterator[Session]:
        with cls.SessionLocal() as session:
            try:
                cls.utils.initialize(session)
                yield session
            finally:
                session.close()

    def get_transactional_session(cls) -> Iterator[Session]:
        with cls.SessionLocal() as session:
            try:
                cls.utils.initialize(session)
                yield session
                session.commit()
            except Exception as e:
                session.rollback()
                raise e
            finally:
                session.close()

    def get_session_scoped(cls) -> scoped_session:
        return scoped_session(cls.SessionLocal)

    def close(cls):
        cls.engine.dispose()


db = PG()



from sqlalchemy.orm.session import Session
from sqlalchemy.orm.query import Query
from sqlalchemy.dialects import postgresql
from sqlalchemy import Select

import logging
import uuid
from typing import Any, List, Optional, Union
from abc import ABC, abstractmethod


class BaseModel:
    id: int

    def __init__(self):
        pass

    def table_name(self) -> str:
        return self.__tablename__

    def to_dict(self) -> dict:
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class PGUtilsBase(ABC):
    def __init__(
        cls, logger: logging.Logger, single_transaction: bool = False, **kwargs
    ):
        cls.session = None
        cls.logger = logger
        cls.single_transaction = single_transaction
        cls.snake_case = kwargs.get("snake_case", False)

    def initialize(cls, session: Session):
        cls.session = session

    @abstractmethod
    def select(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[List[dict], None]:
        pass

    @abstractmethod
    def select_one(cls, session: Session, stmt: Select, **kwargs) -> Union[dict, None]:
        pass

    @abstractmethod
    def select_one_strict(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[BaseModel, Exception]:
        pass

    @abstractmethod
    def check_exists(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[bool, Exception]:
        pass

    @abstractmethod
    def execute(cls, session: Session, stmt: Select) -> Union[bool, None]:
        pass

    @abstractmethod
    def update(
        cls, session: Session, Model: BaseModel, filter_by: dict, values: dict, **kwargs
    ) -> BaseModel:
        pass

    @abstractmethod
    def bulk_update(
        cls, session: Session, model: Any, records: List[dict]
    ) -> Union[bool, None]:
        pass

    @abstractmethod
    def insert(
        cls, session: Session, model, record: dict, **kwargs
    ) -> Union[object, None]:
        pass

    @abstractmethod
    def bulk_insert(
        cls, session: Session, model: Any, records: List[dict], **kwargs
    ) -> List[dict]:
        pass

    @abstractmethod
    def insert_on_conflict(cls, session: Session, model: Any, records: List[dict]):
        pass

    @abstractmethod
    def delete(cls, session: Session, record: BaseModel) -> bool:
        pass

    @abstractmethod
    def delete_by_id(
        cls, session: Session, model: Any, record_id: Union[int, uuid.UUID]
    ) -> bool:
        pass

    @staticmethod
    def __to_snake_case(camel_str: str) -> str:
        """
        Convert a camelCase string to snake_case.

        Parameters:
        camel_str (str): The camelCase string to convert.

        Returns:
        str: The string in snake_case.
        """
        snake_str = camel_str[0].lower()
        for char in camel_str[1:]:
            if char.isupper():
                snake_str += "_"
            snake_str += char.lower()
        return snake_str

    def to_snake_case(cls, results: List[dict]) -> List[dict]:
        """
        Convert all keys in a list of dictionaries from camelCase to snake_case.

        Parameters:
        results (List[Dict[str, any]]): A list of dictionaries with camelCase keys.

        Returns:
        List[Dict[str, any]]: A list of dictionaries with keys in snake_case.
        """
        return [
            {cls.__to_snake_case(key): value for key, value in record.items()}
            for record in results
        ]

    @staticmethod
    def __to_camel_case(snake_str: str) -> str:
        """
        Convert a snake_case string to camelCase.

        Parameters:
        snake_str (str): The snake_case string to convert.

        Returns:
        str: The string in camelCase.
        """
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    def results_to_camel_case(cls, results: List[dict]) -> List[dict]:
        """
        Convert all keys in a list of dictionaries from snake_case to camelCase.

        Parameters:
        results (List[Dict[str, any]]): A list of dictionaries with snake_case keys.

        Returns:
        List[Dict[str, any]]: A list of dictionaries with keys in camelCase.
        """
        return [
            {cls.__to_camel_case(key): value for key, value in record.items()}
            for record in results
        ]

    def print_query(cls, query: Query) -> str:
        """
        Print the query generated by a SQLAlchemy Query object.

        Parameters:
        query (Query): The SQLAlchemy Query object to print.

        Returns:
        str: The query generated by the Query object.
        """
        return str(
            query.statement.compile(
                dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
            )
        )



from pg_alchemy_kit.PG import PG
from pg_alchemy_kit.PGUtils import PGUtils, get_engine_url
from pg_alchemy_kit.cacheStrategies import CachingSession, InMemoryCacheStrategy

__all__ = ["PG", "PGUtils", "CachingSession", "InMemoryCacheStrategy", "get_engine_url"]



from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm.session import Session
from sqlalchemy import create_engine, text, Select
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.query import Query
from sqlalchemy.dialects import postgresql
from sqlalchemy import select

import logging
from typing import Any, List, Optional, Union
import uuid
import os
import warnings


def deprecated(func):
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"Call to deprecated function {func.__name__}. {func.__doc__}",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return func(*args, **kwargs)

    return wrapper


class BaseModel:
    id: int

    def __init__(self):
        pass

    def table_name(self) -> str:
        return self.__tablename__

    def to_dict(self) -> dict:
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class PGUtils:
    def __init__(
        cls, logger: logging.Logger, single_transaction: bool = False, **kwargs
    ):
        cls.session = None
        cls.logger = logger
        cls.single_transaction = single_transaction
        cls.snake_case = kwargs.get("snake_case", False)

    def initialize(cls, session: Session):
        cls.session = session

    @staticmethod
    def wrap_to_json(stmt: Union[str, text]) -> text:
        if type(stmt) == str:
            stmt = stmt.replace(";", "")

        return text(f"SELECT json_agg(t) FROM ({stmt}) t")

    def select(cls, session: Session, sql: str, **kwargs) -> Union[List[dict], None]:
        try:
            params = kwargs.get("params", {})
            to_camel_case = kwargs.get("to_camel_case", False)

            stmt: text = cls.wrap_to_json(sql)
            results = session.execute(stmt, params=params).fetchone()[0]
            if results is None:
                return []

            if to_camel_case:
                results = cls.results_to_camel_case(results)

            return results
        except DBAPIError as e:
            raise e

    def select_orm(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[List[dict], None]:
        try:
            convert_to_dict = kwargs.get("convert_to_dict", False)
            results = session.execute(stmt).scalars().all()
            if results is None:
                return []
            if convert_to_dict:
                return [record.to_dict() for record in results]

            return results

        except DBAPIError as e:
            raise e

    def select_orm_one(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[dict, None]:
        try:
            convert_to_dict = kwargs.get("convert_to_dict", False)
            results: BaseModel = session.execute(stmt).scalars().one()
            if results is None:
                return {}
            if convert_to_dict:
                return results.to_dict()

            return results

        except DBAPIError as e:
            raise e

    def select_orm_one_strict(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[BaseModel, Exception]:
        result: Optional[BaseModel] = session.execute(stmt).scalars().one()
        if result is None:
            raise Exception("No records found")
        return result

    def check_exists_orm(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[bool, Exception]:
        try:
            result: Optional[BaseModel] = session.execute(stmt).scalars().all()

            if result is None:
                return False
            return len(result) > 0

        except DBAPIError as e:
            raise e

    def insert(cls, session: Session, sql: str, params: dict) -> Union[bool, None]:
        try:
            stmt: text = text(sql)
            insert = session.execute(stmt, params=params)
            count = insert.rowcount
            cls.logger.info(f"Inserted {count} rows")

            if not cls.single_transaction:
                session.commit()
            else:
                session.flush()

            return count > 0
        except DBAPIError as e:
            raise e

    def delete(cls, session: Session, sql: str, params: dict) -> Union[bool, None]:
        try:
            stmt: text = text(sql)
            insert = session.execute(stmt, params=params)
            count = insert.rowcount
            cls.logger.info(f"Deleted {count} rows")
            if not cls.single_transaction:
                session.commit()

            return count > 0
        except DBAPIError as e:
            raise e

    def execute(cls, session: Session, sql: str) -> Union[bool, None]:
        try:
            stmt: text = text(sql)
            session.execute(stmt)

            return True
        except DBAPIError as e:
            raise e

    def execute_orm(cls, session: Session, stmt: Select) -> Union[bool, None]:
        try:
            return session.execute(stmt).fetchall()
        except DBAPIError as e:
            raise e

    def query(cls, query: Query) -> Union[bool, None]:
        try:
            return query.all()
        except DBAPIError as e:
            raise e

    # @deprecated
    def update(
        cls, session: Session, model: Any, key_value: dict, update_values: dict
    ) -> Union[bool, None]:
        """This method is deprecated. Use update() instead."""
        try:
            key = list(key_value.keys())[0]
            update_stmt = " , ".join([f"{k} = :{k}" for k in update_values.keys()])

            stmt = text(
                f"UPDATE {model().table_name()} SET {update_stmt} WHERE {key} = :{key}"
            )
            session.execute(stmt, {**key_value, **update_values})

            if not cls.single_transaction:
                session.commit()
            return True
        except DBAPIError as e:
            cls.logger.info(f"Error in update: {e}")
            if not cls.single_transaction:
                session.rollback()
            raise e

    def update_orm(
        cls, session: Session, Model: BaseModel, filter_by: dict, values: dict, **kwargs
    ) -> BaseModel:
        try:
            obj = session.query(Model).filter_by(**filter_by).one()
            to_snake_case = kwargs.get("to_snake_case", cls.snake_case)

            if to_snake_case:
                values = cls.to_snake_case([values])[0]

            for key, value in values.items():
                setattr(obj, key, value)

            if not cls.single_transaction:
                session.commit()

            return obj

        except DBAPIError as e:
            cls.logger.info(f"Error in update: {e}")
            if not cls.single_transaction:
                session.rollback()
            raise e

    def bulk_update_orm(
        cls, session: Session, model: Any, records: List[dict]
    ) -> Union[bool, None]:
        try:
            session.bulk_update_mappings(model, records)
            if not cls.single_transaction:
                session.commit()
            return True
        except DBAPIError as e:
            cls.logger.info(f"Error in bulk_update: {e}")
            if not cls.single_transaction:
                session.rollback()
            raise e

    def insert_orm(
        cls, session: Session, model, record: dict, **kwargs
    ) -> Union[object, None]:
        try:
            to_snake_case = kwargs.get("to_snake_case", cls.snake_case)

            if to_snake_case:
                record = cls.to_snake_case([record])[0]

            obj = model(**record)
            session.add(obj)
            if not cls.single_transaction:
                session.commit()
            else:
                session.flush()
            return obj
        except DBAPIError as e:
            cls.logger.info(f"Error in add_record_sync: {e}")
            session.rollback()
            return None

    def bulk_insert_orm(
        cls, session: Session, model: Any, records: List[dict], **kwargs
    ) -> List[dict]:
        try:
            records_to_insert: List[dict] = [model(**record) for record in records]

            session.add_all(records_to_insert)
            session.flush()  # Flush the records to obtain their IDs
            records: dict = [record.to_dict() for record in records_to_insert]

            if not cls.single_transaction:
                session.commit()

            return records
        except DBAPIError as e:
            cls.session.rollback()
            cls.logger.info(f"Error in add_records_sync: {e}")
            return []

    def insert_orm_on_conflict(
        cls,
        session: Session,
        model: Any,
        records: List[dict],
    ):
        for record in records:
            cls.insert_orm(session, model, record)

    def delete_orm(cls, session: Session, record: BaseModel) -> bool:
        try:
            session.delete(record)
            if not cls.single_transaction:
                session.commit()
            return True
        except DBAPIError as e:
            if not cls.single_transaction:
                session.rollback()
            cls.logger.info(f"Error in remove_records_sync: {e}")
            return False

    def delete_orm_by_id(
        cls, session: Session, model: Any, record_id: Union[int, uuid.UUID]
    ) -> bool:
        try:
            stmt = select(model).where(model.id == record_id)
            record: BaseModel = cls.select_orm_one_strict(session, stmt)
            return cls.delete_orm(session, record)
        except DBAPIError as e:
            cls.logger.info(f"Error in remove_records_sync: {e}")
            return False

    def get_uuid(
        cls,
        session: Session,
        model: Any,
        key_value: dict,
    ) -> uuid.UUID:
        table_name = model().table_name()

        try:
            key = list(key_value.keys())[0]
            stmt = f"SELECT uuid FROM {table_name} WHERE {key} = :{key}"
            return cls.select(session, stmt, key_value)[0]["uuid"]
        except Exception as e:
            cls.logger.info(f"Error in get_uuid: {e}")
            return None

    @staticmethod
    def __to_snake_case(camel_str: str) -> str:
        """
        Convert a camelCase string to snake_case.

        Parameters:
        camel_str (str): The camelCase string to convert.

        Returns:
        str: The string in snake_case.
        """
        snake_str = camel_str[0].lower()
        for char in camel_str[1:]:
            if char.isupper():
                snake_str += "_"
            snake_str += char.lower()
        return snake_str

    def to_snake_case(cls, results: List[dict]) -> List[dict]:
        """
        Convert all keys in a list of dictionaries from camelCase to snake_case.

        Parameters:
        results (List[Dict[str, any]]): A list of dictionaries with camelCase keys.

        Returns:
        List[Dict[str, any]]: A list of dictionaries with keys in snake_case.
        """
        return [
            {cls.__to_snake_case(key): value for key, value in record.items()}
            for record in results
        ]

    @staticmethod
    def __to_camel_case(snake_str: str) -> str:
        """
        Convert a snake_case string to camelCase.

        Parameters:
        snake_str (str): The snake_case string to convert.

        Returns:
        str: The string in camelCase.
        """
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    def results_to_camel_case(cls, results: List[dict]) -> List[dict]:
        """
        Convert all keys in a list of dictionaries from snake_case to camelCase.

        Parameters:
        results (List[Dict[str, any]]): A list of dictionaries with snake_case keys.

        Returns:
        List[Dict[str, any]]: A list of dictionaries with keys in camelCase.
        """
        return [
            {cls.__to_camel_case(key): value for key, value in record.items()}
            for record in results
        ]

    def print_query(cls, query: Query) -> str:
        """
        Print the query generated by a SQLAlchemy Query object.

        Parameters:
        query (Query): The SQLAlchemy Query object to print.

        Returns:
        str: The query generated by the Query object.
        """
        return str(
            query.statement.compile(
                dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
            )
        )


def get_engine(url: str, **kwargs) -> Engine:
    try:
        pool_size = kwargs.get("pool_size", 5)
        max_overflow = kwargs.get("max_overflow", 0)
        pool_pre_ping = kwargs.get("pool_pre_ping", True)
        return create_engine(
            url,
            pool_size=pool_size,
            max_overflow=max_overflow,
            pool_pre_ping=pool_pre_ping,
            **kwargs,
        )
    except DBAPIError as e:
        raise e


def get_engine_url(
    connection_type: str = "postgresql", settings: Any = None, **kwargs
) -> str:
    if settings is not None:
        return f"{connection_type}://{settings.pg_username}:{settings.pg_password}@{settings.pg_host}:{settings.pg_port}/{settings.pg_db}"

    username = kwargs.get("pg_username", os.environ.get("PG_USERNAME"))
    password = kwargs.get("pg_password", os.environ.get("PG_PASSWORD"))
    host = kwargs.get("pg_host", os.environ.get("PG_HOST"))
    port = kwargs.get("pg_port", os.environ.get("PG_PORT"))
    db = kwargs.get("pg_db", os.environ.get("PG_DB"))

    return f"{connection_type}://{username}:{password}@{host}:{port}/{db}"



import uuid
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm.session import Session
from sqlalchemy import select, Select

from typing import Any, List, Optional, Union


from pg_alchemy_kit.PGUtilsBase import PGUtilsBase, BaseModel


class PGUtilsORM(PGUtilsBase):
    def select(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[List[dict], None]:
        try:
            convert_to_dict = kwargs.get("convert_to_dict", False)
            results = session.execute(stmt).scalars().all()
            if results is None:
                return []
            if convert_to_dict:
                return [record.to_dict() for record in results]

            return results

        except DBAPIError as e:
            raise e

    def select_one(cls, session: Session, stmt: Select, **kwargs) -> Union[dict, None]:
        try:
            convert_to_dict = kwargs.get("convert_to_dict", False)
            results: BaseModel = session.execute(stmt).scalars().one()
            if results is None:
                return {}
            if convert_to_dict:
                return results.to_dict()

            return results

        except DBAPIError as e:
            raise e

    def select_one_strict(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[BaseModel, Exception]:
        result: Optional[BaseModel] = session.execute(stmt).scalars().one()
        if result is None:
            raise Exception("No records found")
        return result

    def check_exists(
        cls, session: Session, stmt: Select, **kwargs
    ) -> Union[bool, Exception]:
        try:
            result: Optional[BaseModel] = session.execute(stmt).scalars().all()

            if result is None:
                return False
            return len(result) > 0

        except DBAPIError as e:
            raise e

    def execute(cls, session: Session, stmt: Select) -> Union[bool, None]:
        try:
            return session.execute(stmt).fetchall()
        except DBAPIError as e:
            raise e

    def update(
        cls, session: Session, Model: BaseModel, filter_by: dict, values: dict, **kwargs
    ) -> BaseModel:
        try:
            obj = session.query(Model).filter_by(**filter_by).one()
            to_snake_case = kwargs.get("to_snake_case", cls.snake_case)

            if to_snake_case:
                values = cls.to_snake_case([values])[0]

            for key, value in values.items():
                setattr(obj, key, value)

            if not cls.single_transaction:
                session.commit()

            return obj

        except DBAPIError as e:
            cls.logger.info(f"Error in update: {e}")
            if not cls.single_transaction:
                session.rollback()
            raise e

    def bulk_update(
        cls, session: Session, model: Any, records: List[dict]
    ) -> Union[bool, None]:
        try:
            session.bulk_update_mappings(model, records)
            if not cls.single_transaction:
                session.commit()
            return True
        except DBAPIError as e:
            cls.logger.info(f"Error in bulk_update: {e}")
            if not cls.single_transaction:
                session.rollback()
            raise e

    def insert(
        cls, session: Session, model, record: dict, **kwargs
    ) -> Union[object, None]:
        try:
            to_snake_case = kwargs.get("to_snake_case", cls.snake_case)

            if to_snake_case:
                record = cls.to_snake_case([record])[0]

            obj = model(**record)
            session.add(obj)
            if not cls.single_transaction:
                session.commit()
            else:
                session.flush()
            return obj
        except DBAPIError as e:
            cls.logger.info(f"Error in add_record_sync: {e}")
            session.rollback()
            return None

    def bulk_insert(
        cls, session: Session, model: Any, records: List[dict], **kwargs
    ) -> List[dict]:
        try:
            records_to_insert: List[dict] = [model(**record) for record in records]

            session.add_all(records_to_insert)
            session.flush()  # Flush the records to obtain their IDs
            records: dict = [record.to_dict() for record in records_to_insert]

            if not cls.single_transaction:
                session.commit()

            return records
        except DBAPIError as e:
            cls.session.rollback()
            cls.logger.info(f"Error in add_records_sync: {e}")
            return []

    def insert_on_conflict(
        cls,
        session: Session,
        model: Any,
        records: List[dict],
    ):
        for record in records:
            cls.insert(session, model, record)

    def delete(cls, session: Session, record: BaseModel) -> bool:
        try:
            session.delete(record)
            if not cls.single_transaction:
                session.commit()
            return True
        except DBAPIError as e:
            if not cls.single_transaction:
                session.rollback()
            cls.logger.info(f"Error in remove_records_sync: {e}")
            return False

    def delete_by_id(
        cls, session: Session, model: Any, record_id: Union[int, uuid.UUID]
    ) -> bool:
        try:
            stmt = select(model).where(model.id == record_id)
            record: BaseModel = cls.select_one_strict(session, stmt)
            return cls.delete(session, record)
        except DBAPIError as e:
            cls.logger.info(f"Error in remove_records_sync: {e}")
            return False



import datetime
import pickle
from sqlalchemy.orm import Session
from sqlalchemy import Select, Insert
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.util import _ORMJoin
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound


MINUTE = 10


class CacheMissError(Exception):
    """Raised when the cache miss occurs."""


class CachedResult:
    def __init__(self, data):
        self._data = self._format_data(data)
        self._attributes = {}

    def _format_data(self, data):
        return [v for d in data for v in d.values()]

    def all(self):
        return self._data

    def fetchall(self):
        return self._data

    def fetchone(self):
        return self._data[0] if self._data else None

    def first(self):
        return self._data[0] if self._data else None

    def one(self):
        if len(self._data) == 1:
            return self._data[0]
        elif len(self._data) == 0:
            raise NoResultFound("No result found.")
        else:
            raise MultipleResultsFound("Multiple results found.")

    def scalar(self):
        result = self.one()
        return result[0] if result else None

    def scalars(self):
        return self

    def one_or_none(self):
        try:
            return self.one()
        except MultipleResultsFound:
            raise
        except NoResultFound:
            return None

    def unique(self):
        if len(self._data) == 0:
            raise NoResultFound("No result found.")
        elif len(self._data) > 1:
            raise MultipleResultsFound("Multiple results found.")
        return self._data[0]


class CachingSession(Session):
    def __init__(self, **options):
        self.cache_strategy: InMemoryCacheStrategy = options.pop(
            "cache_strategy", InMemoryCacheStrategy()
        )
        super().__init__(**options)

    def add(self, instance: object, _warn: bool = True) -> None:
        self.cache_strategy.add(self, instance, _warn=_warn)

    def delete(self, instance: object) -> None:
        return self.cache_strategy.delete(self, instance)

    def execute(self, statement, *multiparams, **params):
        if isinstance(statement, Select):
            try:
                return self.cache_strategy.select(
                    self, statement, *multiparams, **params
                )
            except CacheMissError:
                pass

        return super().execute(statement, *multiparams, **params)


class InMemoryCacheStrategy:
    def __init__(self, ttl: int = MINUTE):
        self.cache: dict = {}
        self.expire_times: dict = {}
        self.ttl: int = ttl

    def set_cache(self, cache_key: str, result: any):
        self.cache[cache_key] = pickle.dumps(result)
        self.expire_times[cache_key] = datetime.datetime.now().timestamp() + self.ttl

    def get_raw_data(self, cache_key: str) -> str:
        return self.cache.get(cache_key)

    def load_data(self, raw_data) -> any:
        return pickle.loads(raw_data)

    def _extract_table_name(self, from_clause):
        """Extract table name(s) from a SQLAlchemy clause (table or join)."""

        if hasattr(from_clause, "name"):
            return [from_clause.name]

        elif isinstance(from_clause, _ORMJoin):
            left_names = self._extract_table_name(from_clause.left)
            right_names = self._extract_table_name(from_clause.right)
            return left_names + right_names

        return []

    def extract_table_name(self, statement: Select) -> str:
        return self._extract_table_name(statement.froms[0])[0]

    def check_expired(self, raw_data: any, cache_key: str) -> bool:
        if (
            raw_data
            and self.expire_times.get(cache_key, 0)
            <= datetime.datetime.now().timestamp()
        ):
            print("Expired, removing from cache")
            del self.cache[cache_key]
            raw_data = None

    def select(self, session: Session, statement: Select, *multiparams, **params):
        cache_key = self.generate_cache_key(session, statement)
        raw_data = self.get_raw_data(cache_key)

        self.check_expired(raw_data, cache_key)

        if raw_data is None:
            result = self.__execute(session, statement, *multiparams, **params)
            self.set_cache(cache_key, result)
        else:
            result = self.load_data(raw_data)

        return CachedResult(result)

    def add(self, session: Session, instance: object, _warn: bool = True) -> None:
        self.clear_cache_for_table(session, instance.__table__.name)
        super(CachingSession, session).add(instance, _warn=_warn)

    def delete(self, session: Session, instance: object) -> None:
        self.clear_cache_for_table(session, instance.__table__.name)
        return super(CachingSession, session).delete(instance)

    def clear_cache_for_table(self, session: Session, table_name: str):
        cache_keys = list(filter(lambda x: x.startswith(table_name), self.cache.keys()))

        for cache_key in cache_keys:
            del self.cache[cache_key]

    def __execute(self, session: Session, statement: Select, *multiparams, **params):
        return (
            super(CachingSession, session)
            .execute(statement, *multiparams, **params)
            # .scalars()
            # .all()
            # .fetchall()
            .mappings()
            .all()
        )

    def generate_cache_key(self, session: Session, statement: Select) -> str:
        return f"{self.extract_table_name(statement)}:{self.get_sql_stmt(statement)}"

    def get_sql_stmt(self, statement: Select) -> str:
        "Use the statement's SQL and parameters as the cache key"
        return str(
            statement.compile(
                dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
            )
        )
